import logging
import inspect
import json


class Message(object):
    """ Message generic class """

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps({
            attr: getattr(self, attr)
            for attr in self.__dir__()
            if not attr.startswith('_')
            and not inspect.ismethod(getattr(self, attr))
        }).replace('\n', ' ')

    def __bytes__(self):
        """
        Overrides the bytes() operator, converts the message into
        its JSON-serialized UTF-8-encoded representation
        """
        return str(self).encode('utf-8')

    @classmethod
    def parse(cls, msg):
        """
        Parse a generic message into a key-value dictionary
        Params:
            msg -- Original message - can be a dictionary, a Message,
                   or a string/bytearray, as long as it's valid UTF-8 JSON
        """

        if isinstance(msg, cls):
            msg = str(msg)
        if isinstance(msg, bytes) or isinstance(msg, bytearray):
            msg = msg.decode('utf-8')
        if isinstance(msg, str):
            try:
                msg = json.loads(msg.strip())
            except:
                logging.warning('Invalid JSON message: {}'.format(msg))

        assert isinstance(msg, dict)
        return msg

    @classmethod
    def build(cls, msg):
        """
        Builds a Message object from a dictionary.
        Params:
            msg -- The message as a key-value dictionary, Message object or JSON string
        """
        from platypush.utils import get_message_class_by_type


        msg = cls.parse(msg)
        msgtype = get_message_class_by_type(msg['type'])
        if msgtype != cls: return msgtype.build(msg)

# vim:sw=4:ts=4:et:




import json
import logging
import re

from platypush.config import Config
from platypush.message.event import Event, EventMatchResult
from platypush.message.request import Request
from platypush.utils import get_event_class_by_type


def parse(msg):
    """ Builds a dict given another dictionary or
        a JSON UTF-8 encoded string/bytearray """

    if isinstance(msg, bytes) or isinstance(msg, bytearray):
        msg = msg.decode('utf-8')
    if isinstance(msg, str):
        try:
            msg = json.loads(msg.strip())
        except:
            logging.warning('Invalid JSON message: {}'.format(msg))
            return None

    return msg


class EventCondition(object):
    """ Event hook condition class """

    def __init__(self, type=Event.__class__, **kwargs):
        """
        Rule constructor.
        Params:
            type   -- Class of the event to be built
            kwargs -- Fields rules as a key-value (e.g. source_button=btn_id
                or recognized_phrase='Your phrase')
        """

        self.type = type
        self.args = {}
        self.parsed_args = {}

        for (key, value) in kwargs.items():
            # TODO So far we only allow simple value match. If value is a dict
            # instead, we should allow more a sophisticated attribute matching,
            # e.g. or conditions, in, and other operators.
            self.args[key] = value

    @classmethod
    def build(cls, rule):
        """ Builds a rule given either another EventRule, a dictionary or
            a JSON UTF-8 encoded string/bytearray """

        if isinstance(rule, cls): return rule
        else: rule = parse(rule)
        assert isinstance(rule, dict)

        type = get_event_class_by_type(
            rule.pop('type') if 'type' in rule else 'Event')

        args = {}
        for (key, value) in rule.items():
            args[key] = value

        return cls(type=type, **args)


class EventAction(Request):
    """ Event hook action class. It is a special type of runnable request
        whose fields can be configured later depending on the event context """

    def __init__(self, target=None, action=None, **args):
        if target is None: target=Config.get('device_id')
        super().__init__(target=target, action=action, **args)


    def execute(self, **context):
        event_args = context.pop('event').args if 'event' in context else {}

        for (argname, value) in self.args.items():
            if isinstance(value, str):
                parsed_value = ''
                while value:
                    m = re.match('([^\\\]*)\$([\w\d_-]+)(.*)', value)
                    if m:
                        context_argname = m.group(2)
                        value = m.group(3)
                        if context_argname in context:
                            parsed_value += m.group(1) + context[context_argname]
                        else:
                            parsed_value += m.group(1) + '$' + m.group(2)
                    else:
                        parsed_value += value
                        value = ''

                value = parsed_value

            self.args[argname] = value

        super().execute()


    @classmethod
    def build(cls, action):
        action = super().parse(action)
        action['origin'] = Config.get('device_id')

        if 'target' not in action:
            action['target'] = action['origin']
        return super().build(action)


class EventHook(object):
    """ Event hook class. It consists of one conditionss and
        one or multiple actions to be executed """

    def __init__(self, name, condition=None, actions=[]):
        """ Construtor. Takes a name, a EventCondition object and a list of
            EventAction objects as input """

        self.name = name
        self.condition = EventCondition.build(condition or {})
        self.actions = actions


    @classmethod
    def build(cls, name, hook):
        """ Builds a rule given either another EventRule, a dictionary or
            a JSON UTF-8 encoded string/bytearray """

        if isinstance(hook, cls): return hook
        else: hook = parse(hook)
        assert isinstance(hook, dict)

        condition = EventCondition.build(hook['if']) if 'if' in hook else None
        actions = []
        if 'then' in hook:
            if isinstance(hook['then'], list):
                actions = [EventAction.build(action) for action in hook['then']]
            else:
                actions = [EventAction.build(hook['then'])]

        return cls(name=name, condition=condition, actions=actions)


    def matches_event(self, event):
        """ Returns an EventMatchResult object containing the information
            about the match between the event and this hook """

        return event.matches_condition(self.condition)


    def run(self, event):
        """ Checks the condition of the hook against a particular event and
            runs the hook actions if the condition is met """

        result = self.matches_event(event)

        if result.is_match:
            logging.info('Running hook {} triggered by an event'.format(self.name))

            for action in self.actions:
                action.execute(event=event, **result.parsed_args)


# vim:sw=4:ts=4:et:




import subprocess

from platypush.message.response import Response

from .. import Plugin

class ShellPlugin(Plugin):
    def exec(self, cmd):
        output = None
        errors = []

        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True).decode('utf-8')
        except subprocess.CalledProcessError as e:
            errors = [e.output.decode('utf-8')]

        return Response(output=output, errors=errors)

# vim:sw=4:ts=4:et:




import logging
import json
import time

from kafka import KafkaConsumer, KafkaProducer

from .. import Backend

class KafkaBackend(Backend):
    _conn_retry_secs = 5

    def __init__(self, server, topic, **kwargs):
        super().__init__(**kwargs)

        self.server = server
        self.topic_prefix = topic
        self.topic = self._topic_by_device_id(self.device_id)
        self.producer = None

        logging.getLogger('kafka').setLevel(logging.ERROR)

    def _on_record(self, record):
        if record.topic != self.topic: return

        try:
            msg = json.loads(record.value.decode('utf-8'))
        except Exception as e:
            logging.exception(e)

        logging.debug('Received message on Kafka backend: {}'.format(msg))
        self.on_message(msg)

    def _init_producer(self):
        if not self.producer:
            self.producer = KafkaProducer(bootstrap_servers=self.server)

    def _topic_by_device_id(self, device_id):
        return '{}.{}'.format(self.topic_prefix, device_id)

    def send_message(self, msg):
        target = msg.target
        msg = str(msg).encode('utf-8')

        self._init_producer()
        self.producer.send(self._topic_by_device_id(target), msg)
        self.producer.flush()

    def on_stop(self):
        try:
            if self.producer:
                self.producer.flush()
                self.producer.close()

            if self.consumer:
                self.consumer.close()
        except Exception as e:
            logging.warning('Exception occurred while closing Kafka connection')
            logging.exception(e)

    def run(self):
        super().run()

        self.consumer = KafkaConsumer(self.topic, bootstrap_servers=self.server)
        logging.info('Initialized kafka backend - server: {}, topic: {}'
                     .format(self.server, self.topic))

        try:
            for msg in self.consumer:
                self._on_record(msg)
                if self.should_stop(): break
        except Exception as e:
            logging.warning('Kafka connection error, reconnecting in {} seconds'.
                            format(self._conn_retry_secs))
            logging.exception(e)
            time.sleep(self._conn_retry_secs)

# vim:sw=4:ts=4:et:




import json

from platypush.message import Message

class Response(Message):
    """ Response message class """

    def __init__(self, target=None, origin=None, id=None, output=None, errors=[]):
        """
        Params:
            target -- Target [String]
            origin -- Origin [String]
            output -- Output [String]
            errors -- Errors [List of strings or exceptions]
                id -- Message ID this response refers to
        """

        self.target = target
        self.output = output
        self.errors = errors
        self.origin = origin
        self.id = id

    def is_error(self):
        """ Returns True if the respopnse has errors """
        return len(self.errors) != 0

    @classmethod
    def build(cls, msg):
        msg = super().parse(msg)
        args = {
            'target' : msg['target'],
            'output' : msg['response']['output'],
            'errors' : msg['response']['errors'],
        }

        if 'id' in msg: args['id'] = msg['id']
        if 'origin' in msg: args['origin'] = msg['origin']
        return cls(**args)


    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps({
            'id'         : self.id,
            'type'       : 'response',
            'target'     : self.target if hasattr(self, 'target') else None,
            'origin'     : self.origin if hasattr(self, 'origin') else None,
            'response'   : {
                'output' : self.output,
                'errors' : self.errors,
            },
        })

# vim:sw=4:ts=4:et:




import subprocess
import urllib.parse

from platypush.message.response import Response

from .. import Plugin

class TtsPlugin(Plugin):
    """ Default Text-to-Speech plugin. It leverages Google Translate and
        requires mplayer """

    def __init__(self, lang='en-gb'):
        super().__init__()
        self.lang=lang

    def say(self, phrase, lang=None):
        if lang is None: lang=self.lang
        output = None
        errors = []
        cmd = ['mplayer -ao alsa -really-quiet -noconsolecontrols ' +
               '"http://translate.google.com/translate_tts?{}"'
               .format(urllib.parse.urlencode({
                   'ie'     : 'UTF-8',
                   'client' : 'tw-ob',
                   'tl'     : lang,
                   'q'      : phrase,
                }))]

        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True).decode('utf-8')
        except subprocess.CalledProcessError as e:
            errors = [e.output.decode('utf-8')]

        return Response(output=output, errors=errors)

# vim:sw=4:ts=4:et:




import sys

from . import Pusher

opts = Pusher.parse_build_args(sys.argv[1:])
pusher = Pusher(config_file=opts.config, backend=opts.backend)

if opts.type == 'event':
    pusher.send_event(target=opts.target, type=opts.event, **opts.args)
else:
    pusher.send_request(target=opts.target, action=opts.action, timeout=opts.timeout, **opts.args)


# vim:sw=4:ts=4:et:




import logging
import time

from phue import Bridge
from platypush.message.response import Response

from .. import LightPlugin

class LightHuePlugin(LightPlugin):
    """ Philips Hue lights plugin """

    MAX_BRI = 255
    MAX_SAT = 255
    MAX_HUE = 65535

    def __init__(self, bridge, lights=None, groups=None):
        """
        Constructor
        Params:
            bridge -- Bridge address or hostname
            lights -- Lights to be controlled (default: all)
            groups -- Groups to be controlled (default: all)
        """

        super().__init__()

        self.bridge_address = bridge
        self.bridge = None
        logging.info('Initializing Hue lights plugin - bridge: "{}"'.
                     format(self.bridge_address))

        self.connect()
        self.lights = []; self.groups = []

        if lights:
            self.lights = lights
        elif groups:
            self.groups = groups
            self._expand_groups()
        else:
            self.lights = [l.name for l in self.bridge.lights]

        logging.info('Configured lights: "{}"'. format(self.lights))

    def _expand_groups(self):
        groups = [g for g in self.bridge.groups if g.name in self.groups]
        for g in groups:
            self.lights.extend([l.name for l in g.lights])

    def connect(self):
        # Lazy init
        if not self.bridge:
            self.bridge = Bridge(self.bridge_address)
            logging.info('Bridge connected')

            self.get_scenes()

            # uncomment these lines if you're running huectrl for
            # the first time and you need to pair it to the switch

            # self.bridge.connect()
            # self.bridge.get_api()
        else:
            logging.info('Bridge already connected')


    def get_scenes(self):
        scenes = [s.name for s in self.bridge.scenes]
        # TODO Expand it with custom scenes specified in config.yaml as in #14

    def _exec(self, attr, *args, **kwargs):
        try:
            self.connect()
        except Exception as e:
            # Reset bridge connection
            self.bridge = None
            raise e

        lights = []; groups = []
        if 'lights' in kwargs and kwargs['lights']:
            lights = kwargs['lights'].split(',') \
                if isinstance(lights, str) else kwargs['lights']
        elif 'groups' in kwargs and kwargs['lights']:
            groups = kwargs['groups'].split(',') \
                if isinstance(groups, str) else kwargs['groups']
        else:
            lights = self.lights
            groups = self.groups

        try:
            if attr == 'scene':
                self.bridge.run_scene(groups[0], kwargs['name'])
            elif groups:
                self.bridge.set_group(groups, attr, *args)
            elif lights:
                self.bridge.set_light(lights, attr, *args)
        except Exception as e:
            # Reset bridge connection
            self.bridge = None
            raise e

        return Response(output='ok')

    def on(self, lights=[], groups=[]):
        return self._exec('on', True, lights=lights, groups=groups)

    def off(self, lights=[], groups=[]):
        return self._exec('on', False, lights=lights, groups=groups)

    def bri(self, value, lights=[], groups=[]):
        return self._exec('bri', int(value) % (self.MAX_BRI+1),
                      lights=lights, groups=groups)

    def sat(self, value, lights=[], groups=[]):
        return self._exec('sat', int(value) % (self.MAX_SAT+1),
                      lights=lights, groups=groups)

    def hue(self, value, lights=[], groups=[]):
        return self._exec('hue', int(value) % (self.MAX_HUE+1),
                      lights=lights, groups=groups)

    def scene(self, name, lights=[], groups=[]):
        return self._exec('scene', name=name, lights=lights, groups=groups)


# vim:sw=4:ts=4:et:




import argparse
import logging
import sys
import traceback

from threading import Thread

from .bus import Bus
from .config import Config
from .context import register_backends
from .event.processor import EventProcessor
from .message.event import Event, StopEvent
from .message.request import Request
from .message.response import Response
from .procedure import Procedure

__author__ = 'Fabio Manganiello <blacklight86@gmail.com>'
__version__ = '0.6.4'

#-----------#

class Daemon(object):
    """ Main class for the Platypush daemon """

    """ Configuration file (default: either ~/.config/platypush/config.yaml or
        /etc/platypush/config.yaml) """
    config_file = None

    """ Application bus. It's an internal queue where:
        - backends will post the messages they receive
        - plugins will post the responses they process """
    bus = None

    """ backend_name => backend_obj map """
    backends = None

    """ number of executions retries before a request fails """
    n_tries = 2

    def __init__(self, config_file=None, requests_to_process=None):
        """ Constructor
        Params:
            config_file -- Configuration file override (default: None)
            requests_to_process -- Exit after processing the specified number
                                   of requests (default: None, loop forever)
        """

        self.config_file = config_file
        self.event_processor = EventProcessor()
        self.requests_to_process = requests_to_process
        self.processed_requests = 0

        Config.init(self.config_file)
        logging.basicConfig(level=Config.get('logging'), stream=sys.stdout)

    @classmethod
    def build_from_cmdline(cls, args):
        """ Build the app from command line arguments.
        Params:
            args -- Your sys.argv[1:] [List of strings]
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--config', '-c', dest='config', required=False,
                            default=None, help=cls.config_file.__doc__)

        opts, args = parser.parse_known_args(args)
        return cls(config_file=opts.config)

    def on_message(self):
        """ Default message handler """

        def _f(msg):
            """ on_message closure
            Params:
                msg -- platypush.message.Message instance """

            if isinstance(msg, Request):
                msg.execute(n_tries=self.n_tries)
                self.processed_requests += 1
                if self.requests_to_process \
                        and self.processed_requests >= self.requests_to_process:
                    self.stop_app()
            elif isinstance(msg, Response):
                logging.info('Received response: {}'.format(msg))
            elif isinstance(msg, StopEvent) and msg.targets_me():
                logging.info('Received STOP event: {}'.format(msg))
                self.stop_app()
            elif isinstance(msg, Event):
                logging.info('Received event: {}'.format(msg))
                self.event_processor.process_event(msg)

        return _f


    def stop_app(self):
        """ Stops the backends and the bus """
        for backend in self.backends.values():
            backend.stop()
        self.bus.stop()


    def start(self):
        """ Start the daemon """
        self.bus = Bus(on_message=self.on_message())

        # Initialize the backends and link them to the bus
        self.backends = register_backends(bus=self.bus, global_scope=True)

        # Start the backend threads
        for backend in self.backends.values():
            backend.start()

        # Poll for messages on the bus
        try:
            self.bus.poll()
        except KeyboardInterrupt as e:
            logging.info('SIGINT received, terminating application')
        finally:
            self.stop_app()


# vim:sw=4:ts=4:et:




import os
import sys
import signal
import logging
import threading

from enum import Enum
from queue import Queue

from platypush.config import Config
from platypush.message.event import Event, StopEvent

class Bus(object):
    """ Main local bus where the daemon will listen for new messages """

    def __init__(self, on_message=None):
        self.bus = Queue()
        self.on_message = on_message
        self.thread_id = threading.get_ident()

    def post(self, msg):
        """ Sends a message to the bus """
        self.bus.put(msg)

    def get(self):
        """ Reads one message from the bus """
        return self.bus.get()

    def stop(self):
        """ Stops the bus by sending a STOP event """
        evt = StopEvent(target=Config.get('device_id'),
                        origin=Config.get('device_id'),
                        thread_id=self.thread_id)

        self.bus.put(evt)

    def poll(self):
        """
        Reads messages from the bus until either stop event message or KeyboardInterrupt
        """

        if not self.on_message:
            logging.warning('No message handlers installed, cannot poll')
            return

        stop=False
        while not stop:
            msg = self.get()
            self.on_message(msg)

            if isinstance(msg, StopEvent) and msg.targets_me():
                logging.info('Received STOP event on the bus')
                stop=True


# vim:sw=4:ts=4:et:




import importlib
import logging

from enum import Enum
from threading import Timer
from time import time

from platypush.backend import Backend
from platypush.message.event.button.flic import FlicButtonEvent

from .fliclib.fliclib import FlicClient, ButtonConnectionChannel, ClickType


class ButtonFlicBackend(Backend):
    _long_press_timeout = 0.3
    _btn_timeout = 0.5
    ShortPressEvent = "ShortPressEvent"
    LongPressEvent = "LongPressEvent"

    def __init__(self, server, long_press_timeout=_long_press_timeout,
                 btn_timeout=_btn_timeout, **kwargs):
        super().__init__(**kwargs)

        self.server = server
        self.client = FlicClient(server)
        self.client.get_info(self._received_info())
        self.client.on_new_verified_button = self._got_button()

        self._long_press_timeout = long_press_timeout
        self._btn_timeout = btn_timeout
        self._btn_timer = None
        self._btn_addr = None
        self._down_pressed_time = None
        self._cur_sequence = []

        logging.info('Initialized Flic buttons backend on {}'.format(self.server))

    def _got_button(self):
        def _f(bd_addr):
            cc = ButtonConnectionChannel(bd_addr)
            cc.on_button_up_or_down = \
                lambda channel, click_type, was_queued, time_diff: \
                self._on_event()(bd_addr, channel, click_type, was_queued, time_diff)
            self.client.add_connection_channel(cc)

        return _f

    def _received_info(self):
        def _f(items):
            for bd_addr in items["bd_addr_of_verified_buttons"]:
                self._got_button()(bd_addr)
        return _f

    def _on_btn_timeout(self):
        def _f():
            logging.info('Flic event triggered from {}: {}'.format(
                self._btn_addr, self._cur_sequence))

            self.bus.post(FlicButtonEvent(
                btn_addr=self._btn_addr, sequence=self._cur_sequence))

            self._cur_sequence = []

        return _f

    def _on_event(self):
        def _f(bd_addr, channel, click_type, was_queued, time_diff):
            if was_queued:
                return

            if self._btn_timer:
                self._btn_timer.cancel()

            if click_type == ClickType.ButtonDown:
                self._down_pressed_time = time()
                return

            btn_event = self.ShortPressEvent
            if self._down_pressed_time:
                if time() - self._down_pressed_time >= self._long_press_timeout:
                    btn_event = self.LongPressEvent
                self._down_pressed_time = None

            self._cur_sequence.append(btn_event)

            self._btn_addr = bd_addr
            self._btn_timer = Timer(self._btn_timeout, self._on_btn_timeout())
            self._btn_timer.start()

        return _f

    def send_message(self, msg):
        pass

    def run(self):
        super().run()

        self.client.handle_events()


# vim:sw=4:ts=4:et:




import logging
import json

from multiprocessing import Process
from flask import Flask, abort, jsonify, request

from platypush.message import Message
from platypush.message.request import Request

from .. import Backend


class HttpBackend(Backend):
    """ Example interaction with the HTTP backend to make requests:
        $ curl -XPOST \
            -d 'msg={"type":"request","target":"volta","action":"tts.say","args": {"phrase":"This is a test"}}' \
            http://localhost:8008 """

    def __init__(self, port=8008, token=None, **kwargs):
        super().__init__(**kwargs)
        self.port = port
        self.token = token


    def send_message(self, msg):
        raise NotImplementedError('Use cURL or any HTTP client to query the HTTP backend')


    def _start_server(self):
        def app_main():
            app = Flask(__name__)

            @app.route('/', methods=['POST'])
            def index():
                args = { k:v for (k,v) in request.form.items() }

                if self.token:
                    if 'token' not in args or args['token'] != self.token:
                        abort(401)

                if 'msg' not in args:
                    abort(400)

                msg = Message.build(args['msg'])
                logging.debug('Received message on HTTP backend: {}'.format(msg))

                if isinstance(msg, Request):
                    response = msg.execute(async=False)
                    return str(response)

                return jsonify({ 'status': 'ok' })

            app.run(debug=True, host='0.0.0.0', port=self.port)
            logging.info('Initialized HTTP backend on port {}'.format(self.port))

        return app_main

    def run(self):
        super().run()
        self.server_proc = Process(target=self._start_server())
        self.server_proc.start()
        self.server_proc.join()


# vim:sw=4:ts=4:et:




import logging
import json

from ouimeaux.environment import Environment, UnknownDevice
from platypush.message.response import Response

from .. import SwitchPlugin

class SwitchWemoPlugin(SwitchPlugin):
    def __init__(self, discovery_seconds=3):
        super().__init__()

        self.discovery_seconds=discovery_seconds
        self.env = Environment()
        self.env.start()
        self.refresh_devices()

    def refresh_devices(self):
        logging.info('Starting WeMo discovery')
        self.env.discover(seconds=self.discovery_seconds)
        self.devices = self.env.devices

    def _exec(self, method, device, *args, **kwargs):
        if device not in self.devices:
            self.refresh_devices()

        if device not in self.devices:
            raise RuntimeError('Device {} not found'.format(device))

        logging.info('{} -> {}'.format(device, method))
        dev = self.devices[device]
        getattr(dev, method)(*args, **kwargs)

        resp = {'device': device, 'state': dev.get_state()}
        return Response(output=json.dumps(resp))

    def on(self, device):
        return self._exec('on', device)

    def off(self, device):
        return self._exec('off', device)

    def toggle(self, device):
        return self._exec('toggle', device)


# vim:sw=4:ts=4:et:




from .. import Plugin

class LightPlugin(Plugin):
    def on(self):
        raise NotImplementedError()

    def off(self):
        raise NotImplementedError()

    def toggle(self):
        raise NotImplementedError()

    def status(self):
        raise NotImplementedError()


# vim:sw=4:ts=4:et:




import importlib
import logging
import sys
import threading

from threading import Thread

from platypush.bus import Bus
from platypush.config import Config
from platypush.utils import get_message_class_by_type, set_timeout, clear_timeout
from platypush.message import Message
from platypush.message.event import Event, StopEvent
from platypush.message.request import Request
from platypush.message.response import Response

class Backend(Thread):
    """ Parent class for backends """

    _default_response_timeout = 5

    def __init__(self, bus=None, **kwargs):
        """
        Params:
            bus    -- Reference to the Platypush bus where the requests and the
                      responses will be posted [Bus]
            kwargs -- key-value configuration for this backend [Dict]
        """

        # If no bus is specified, create an internal queue where
        # the received messages will be pushed
        self.bus = bus or Bus()
        self.device_id = Config.get('device_id')
        self.thread_id = None
        self._stop = False
        self._kwargs = kwargs

        # Internal-only, we set the request context on a backend if that
        # backend is intended to react for a response to a specific request
        self._request_context = kwargs['_req_ctx'] if '_req_ctx' in kwargs \
            else None

        Thread.__init__(self)
        logging.basicConfig(stream=sys.stdout, level=Config.get('logging')
                            if 'logging' not in kwargs
                            else getattr(logging, kwargs['logging']))


    def on_message(self, msg):
        """
        Callback when a message is received on the backend.
        It parses and posts the message on the main bus.
        It should be called by the derived classes whenever
        a new message should be processed.

        Params:
            msg -- The message. It can be either a key-value
                   dictionary, a platypush.message.Message
                   object, or a string/byte UTF-8 encoded string
        """
        msg = Message.build(msg)

        if not getattr(msg, 'target') or msg.target != self.device_id:
            return  # Not for me

        logging.debug('Message received on the {} backend: {}'.format(
            self.__class__.__name__, msg))

        if self._is_expected_response(msg):
            # Expected response, trigger the response handler
            clear_timeout()
            self._request_context['on_response'](msg)
            self.stop()
            return

        if isinstance(msg, StopEvent) and msg.targets_me():
            logging.info('Received STOP event on {}'.format(self.__class__.__name__))
            self._stop = True
        else:
            msg.backend = self   # Augment message to be able to process responses
            self.bus.post(msg)


    def _is_expected_response(self, msg):
        """ Internal only - returns true if we are expecting for a response
            and msg is that response """

        return  self._request_context \
            and isinstance(msg, Response) \
            and msg.id == self._request_context['request'].id


    def _get_backend_config(self):
        config_name = 'backend.' + self.__class__.__name__.split('Backend')[0].lower()
        return Config.get(config_name)


    def _setup_response_handler(self, request, on_response, response_timeout):
        def _timeout_hndl():
            raise RuntimeError('Timed out while waiting for a response from {}'.
                               format(request.target))

        req_ctx = {
            'request': request,
            'on_response': on_response,
            'response_timeout': response_timeout,
        }

        resp_backend = self.__class__(bus=self.bus, _req_ctx=req_ctx,
                                     **self._get_backend_config(), **self._kwargs)

        # Set the response timeout
        if response_timeout:
            set_timeout(seconds=response_timeout, on_timeout=_timeout_hndl)

        resp_backend.start()


    def send_event(self, event, **kwargs):
        """
        Send an event message on the backend
        Params:
            event -- The request, either a dict, a string/bytes UTF-8 JSON,
                     or a platypush.message.event.Event object.
        """

        event = Event.build(event)
        assert isinstance(event, Event)

        event.origin = self.device_id
        if not hasattr(event, 'target'):
            event.target = self.device_id

        self.send_message(event, **kwargs)


    def send_request(self, request, on_response=None,
                     response_timeout=_default_response_timeout, **kwargs):
        """
        Send a request message on the backend
        Params:
            request     -- The request, either a dict, a string/bytes UTF-8 JSON,
                           or a platypush.message.request.Request object.

            on_response -- Response handler, takes a platypush.message.response.Response
                           as argument. If set, the method will wait for a
                           response before exiting (default: None)
            response_timeout -- If on_response is set, the backend will raise
                                an exception if the response isn't received
                                within this number of seconds (default: 5)
        """

        request = Request.build(request)
        assert isinstance(request, Request)

        request.origin = self.device_id

        if on_response and response_timeout != 0:
            self._setup_response_handler(request, on_response, response_timeout)

        self.send_message(request, **kwargs)


    def send_response(self, response, request, **kwargs):
        """
        Send a response message on the backend
        Params:
            response -- The response, either a dict, a string/bytes UTF-8 JSON,
                        or a platypush.message.response.Response object
            request  -- Associated request, used to set the response parameters
                        that will link them
        """

        response = Response.build(response)
        assert isinstance(response, Response)
        assert isinstance(request, Request)

        response.id = request.id
        response.target = request.origin
        response.origin = self.device_id

        self.send_message(response, **kwargs)


    def send_message(self, msg, **kwargs):
        """
        Sends a platypush.message.Message to a node.
        To be implemented in the derived classes.
        Always call send_request or send_response instead of send_message directly

        Param:
            msg -- The message
        """

        raise NotImplementedError("send_message should be implemented in a derived class")


    def run(self):
        """ Starts the backend thread. To be implemented in the derived classes """
        self.thread_id = threading.get_ident()

    def on_stop(self):
        """ Callback invoked when the process stops """
        pass

    def stop(self):
        """ Stops the backend thread by sending a STOP event on its bus """
        def _async_stop():
            evt = StopEvent(target=self.device_id, origin=self.device_id,
                            thread_id=self.thread_id)

            self.send_message(evt)
            self.on_stop()

        Thread(target=_async_stop).start()

    def should_stop(self):
        return self._stop


# vim:sw=4:ts=4:et:




import sys
import logging

from platypush.config import Config
from platypush.message.response import Response

class Plugin(object):
    """ Base plugin class """

    def __init__(self, **kwargs):
        logging.basicConfig(stream=sys.stdout, level=Config.get('logging')
                            if 'logging' not in kwargs
                            else getattr(logging, kwargs['logging']))

    def run(self, method, *args, **kwargs):
        return getattr(self, method)(*args, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.message.event import Event, EventMatchResult


class PingEvent(Event):
    """ Ping event """

    def __init__(self, message=None, *args, **kwargs):
        super().__init__(message=message, *args, **kwargs)


# vim:sw=4:ts=4:et:




import functools
import importlib
import logging

from ..config import Config

# Map: backend_name -> backend_instance
backends = {}

# Map: plugin_name -> plugin_instance
plugins = {}

def register_backends(bus=None, global_scope=False, **kwargs):
    """ Initialize the backend objects based on the configuration and returns
        a name -> backend_instance map.
    Params:
        bus -- If specific (it usually should), the messages processed by the
            backends will be posted on this bus.

        kwargs -- Any additional key-value parameters required to initialize the backends
        """

    if global_scope:
        global backends
    else:
        backends = {}

    for (name, cfg) in Config.get_backends().items():
        module = importlib.import_module('platypush.backend.' + name)

        # e.g. backend.pushbullet main class: PushbulletBackend
        cls_name = functools.reduce(
            lambda a,b: a.title() + b.title(),
            (module.__name__.title().split('.')[2:])
        ) + 'Backend'

        try:
            b = getattr(module, cls_name)(bus=bus, **cfg, **kwargs)
            backends[name] = b
        except AttributeError as e:
            logging.warning('No such class in {}: {}'.format(
                module.__name__, cls_name))
            raise RuntimeError(e)

    return backends

def get_backend(name):
    """ Returns the backend instance identified by name if it exists """

    global backends
    return backends[name]


def get_plugin(plugin_name, reload=False):
    """ Registers a plugin instance by name if not registered already, or
        returns the registered plugin instance"""
    global plugins

    if plugin_name in plugins and not reload:
        return plugins[plugin_name]

    try:
        plugin = importlib.import_module('platypush.plugins.' + plugin_name)
    except ModuleNotFoundError as e:
        logging.warning('No such plugin: {}'.format(plugin_name))
        raise RuntimeError(e)

    # e.g. plugins.music.mpd main class: MusicMpdPlugin
    cls_name = functools.reduce(
        lambda a,b: a.title() + b.title(),
        (plugin_name.title().split('.'))
    ) + 'Plugin'

    plugin_conf = Config.get_plugins()[plugin_name] \
        if plugin_name in Config.get_plugins() else {}

    try:
        plugin_class = getattr(plugin, cls_name)
        plugin = plugin_class(**plugin_conf)
        plugins[plugin_name] = plugin
    except AttributeError as e:
        logging.warning('No such class in {}: {}'.format(plugin_name, cls_name))
        raise RuntimeError(e)

    plugins[plugin_name] = plugin
    return plugin

def register_plugin(name, plugin, **kwargs):
    """ Registers a plugin instance by name """
    global plugins

# vim:sw=4:ts=4:et:




import logging
import json
import os
import subprocess
import time

import google.oauth2.credentials

from google.assistant.library import Assistant
from google.assistant.library.event import EventType
from google.assistant.library.file_helpers import existing_file

from platypush.backend import Backend
from platypush.message.event.assistant import \
    ConversationStartEvent, ConversationEndEvent, SpeechRecognizedEvent

class AssistantGoogleBackend(Backend):
    """ Class for the Google Assistant backend. It creates and event source
        that posts recognized phrases on the main bus """

    def __init__(self, credentials_file=os.path.join(
            os.path.expanduser('~/.config'),
            'google-oauthlib-tool', 'credentials.json'), **kwargs):
        """ Params:
            credentials_file -- Path to the Google OAuth credentials file
                (default: ~/.config/google-oauthlib-tool/credentials.json)
        """

        super().__init__(**kwargs)
        self.credentials_file = credentials_file
        self.assistant = None

        with open(self.credentials_file, 'r') as f:
            self.credentials = google.oauth2.credentials.Credentials(token=None,
                                                                     **json.load(f))
        logging.info('Initialized Google Assistant backend')

    def _process_event(self, event):
        logging.info('Received assistant event: {}'.format(event))

        if event.type == EventType.ON_CONVERSATION_TURN_STARTED:
            self.bus.post(ConversationStartEvent())
        elif event.type == EventType.ON_CONVERSATION_TURN_FINISHED:
            self.bus.post(ConversationEndEvent())
        elif event.type == EventType.ON_RECOGNIZING_SPEECH_FINISHED:
            phrase = event.args['text'].lower().strip()
            logging.info('Speech recognized: {}'.format(phrase))
            self.bus.post(SpeechRecognizedEvent(phrase=phrase))


    def start_conversation(self):
        if self.assistant: self.assistant.start_conversation()


    def stop_conversation(self):
        if self.assistant: self.assistant.stop_conversation()


    def send_message(self, msg):
        # Can't send a message on an event source, ignoring
        # TODO Make a class for event sources like these. Event sources
        # would be a subset of the backends which can fire events on the bus
        # but not receive requests or process responses.
        pass

    def run(self):
        super().run()

        with Assistant(self.credentials) as assistant:
            self.assistant = assistant
            for event in assistant.start():
                self._process_event(event)


# vim:sw=4:ts=4:et:




import json
import logging
import random
import traceback

from threading import Thread

from platypush.context import get_plugin
from platypush.message import Message
from platypush.message.response import Response
from platypush.utils import get_module_and_method_from_action

class Request(Message):
    """ Request message class """

    def __init__(self, target, action, origin=None, id=None, backend=None, args=None):
        """
        Params:
            target -- Target node [String]
            action -- Action to be executed (e.g. music.mpd.play) [String]
            origin -- Origin node [String]
                id -- Message ID, or None to get it auto-generated
           backend -- Backend connected to the request, where the response will be delivered
              args -- Additional arguments for the action [Dict]
        """

        self.id      = id if id else self._generate_id()
        self.target  = target
        self.action  = action
        self.origin  = origin
        self.args    = args if args else {}
        self.backend = backend

    @classmethod
    def build(cls, msg):
        msg = super().parse(msg)
        args = {
            'target' : msg['target'],
            'action' : msg['action'],
            'args'   : msg['args'] if 'args' in msg else {},
        }

        args['id'] = msg['id'] if 'id' in msg else cls._generate_id()
        if 'origin' in msg: args['origin'] = msg['origin']
        return cls(**args)

    @staticmethod
    def _generate_id():
        id = ''
        for i in range(0,16):
            id += '%.2x' % random.randint(0, 255)
        return id


    def _execute_procedure(self, *args, **kwargs):
        from config import Config

        logging.info('Executing procedure request: {}'.format(procedure))
        proc_name = self.action.split('.')[-1]
        proc_config = Config.get_procedures()[proc_name]
        proc = Procedure.build(name=proc_name, requests=proc_config, backend=self.backend, id=self.id)
        proc.execute(*args, **kwargs)


    def execute(self, n_tries=1, async=True):
        """
        Execute this request and returns a Response object
        Params:
            n_tries -- Number of tries in case of failure before raising a RuntimeError
            async   -- If True, the request will be run asynchronously and the
                       response posted on the bus when available (default),
                       otherwise the current thread will wait for the response
                       to be returned synchronously.
        """

        def _thread_func(n_tries):
            if self.action.startswith('procedure.'):
                return self._execute_procedure(n_tries=n_tries)

            (module_name, method_name) = get_module_and_method_from_action(self.action)
            plugin = get_plugin(module_name)

            try:
                # Run the action
                response = plugin.run(method=method_name, **self.args)
                if response and response.is_error():
                    raise RuntimeError('Response processed with errors: {}'.format(response))

                logging.info('Processed response from plugin {}: {}'.
                                format(plugin, response))
            except Exception as e:
                # Retry mechanism
                response = Response(output=None, errors=[str(e), traceback.format_exc()])
                logging.exception(e)
                if n_tries:
                    logging.info('Reloading plugin {} and retrying'.format(module_name))
                    get_plugin(module_name, reload=True)
                    _thread_func(n_tries-1)
                    return
            finally:
                if async:
                    # Send the response on the backend
                    if self.backend and self.origin:
                        self.backend.send_response(response=response, request=self)
                    else:
                        logging.info('Response whose request has no ' +
                                    'origin attached: {}'.format(response))
                else:
                    return response

        if async:
            Thread(target=_thread_func, args=(n_tries,)).start()
        else:
            return _thread_func(n_tries)


    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps({
            'type'   : 'request',
            'target' : self.target,
            'action' : self.action,
            'args'   : self.args,
            'origin' : self.origin if hasattr(self, 'origin') else None,
            'id'     : self.id if hasattr(self, 'id') else None,
        })


# vim:sw=4:ts=4:et:




import importlib
import logging
import signal

from platypush.config import Config
from platypush.message import Message


def get_module_and_method_from_action(action):
    """ Input  : action=music.mpd.play
        Output : ('music.mpd', 'play') """

    tokens = action.split('.')
    module_name = str.join('.', tokens[:-1])
    method_name = tokens[-1:][0]
    return (module_name, method_name)


def get_message_class_by_type(msgtype):
    """ Gets the class of a message type given as string """

    try:
        module = importlib.import_module('platypush.message.' + msgtype)
    except ModuleNotFoundError as e:
        logging.warning('Unsupported message type {}'.format(msgtype))
        raise RuntimeError(e)

    cls_name = msgtype[0].upper() + msgtype[1:]

    try:
        msgclass = getattr(module, cls_name)
    except AttributeError as e:
        logging.warning('No such class in {}: {}'.format(
            module.__name__, cls_name))
        raise RuntimeError(e)

    return msgclass


def get_event_class_by_type(type):
    """ Gets an event class by type name """
    event_module = importlib.import_module('.'.join(type.split('.')[:-1]))
    return getattr(event_module, type.split('.')[-1])


def set_timeout(seconds, on_timeout):
    """
    Set a function to be called if timeout expires without being cleared.
    It only works on the main thread.

    Params:
        seconds    -- Timeout in seconds
        on_timeout -- Function invoked on timeout unless clear_timeout is called before
    """

    def _sighandler(signum, frame):
        on_timeout()

    signal.signal(signal.SIGALRM, _sighandler)
    signal.alarm(seconds)


def clear_timeout():
    """ Clear any previously set timeout """
    signal.alarm(0)


# vim:sw=4:ts=4:et:




import mpd

from platypush.message.response import Response

from .. import MusicPlugin

class MusicMpdPlugin(MusicPlugin):
    def __init__(self, host, port):
        """
        Constructor
        Params:
            host -- MPD host
            port -- MPD port
        """

        self.host = host
        self.port = port
        self.client = mpd.MPDClient(use_unicode=True)
        self.client.connect(self.host, self.port)

    def _exec(self, method, *args, **kwargs):
        getattr(self.client, method)(*args, **kwargs)
        return self.status()

    def play(self, resource=None):
        if resource:
            self.clear()
            self.add(resource)
        return self._exec('play')

    def pause(self):
        status = self.status().output['state']
        if status == 'play': return self._exec('pause')
        else: return self._exec('play')

    def stop(self):
        return self._exec('stop')

    def play_or_stop(self):
        status = self.status().output['state']
        if status == 'play': return self._exec('stop')
        else: return self._exec('play')

    def next(self):
        return self._exec('next')

    def previous(self):
        return self._exec('previous')

    def setvol(self, vol):
        return self._exec('setvol', vol)

    def volup(self, delta=10):
        volume = int(self.status().output['volume'])
        new_volume = volume+delta
        if new_volume <= 100:
            self.setvol(str(new_volume))
        return self.status()

    def voldown(self, delta=10):
        volume = int(self.status().output['volume'])
        new_volume = volume-delta
        if new_volume >= 0:
            self.setvol(str(new_volume))
        return self.status()

    def add(self, resource):
        return self._exec('add', resource)

    def load(self, playlist):
        self._exec('load', playlist)
        return self.play()

    def clear(self):
        return self._exec('clear')

    def forward(self):
        return self._exec('seekcur', '+15')

    def back(self):
        return self._exec('seekcur', '-15')

    def status(self):
        return Response(output=self.client.status())

# vim:sw=4:ts=4:et:




import logging
import sys

from ..hook import EventHook

from platypush.config import Config


class EventProcessor(object):
    """ Event processor class. Checks an event against the configured
        rules and executes any matching event hooks """

    def __init__(self, hooks=None, **kwargs):
        """
        Params:
            hooks -- List of event hooks (default: any entry in the config
            named as event.hook.<hook_name> """

        if hooks is None: hooks = Config.get_event_hooks()

        self.hooks = []
        for (name, hook) in hooks.items():
            h = EventHook.build(name=name, hook=hook)
            self.hooks.append(h)


    def process_event(self, event):
        """ Processes an event and runs the matched hooks with the highest score """

        matched_hooks = []
        max_score = -sys.maxsize

        for hook in self.hooks:
            match = hook.matches_event(event)
            if match.is_match:
                if match.score > max_score:
                    matched_hooks = [hook]
                elif match.score == max_score:
                    matched_hooks.append(hook)

        for hook in matched_hooks:
            hook.run(event)


# vim:sw=4:ts=4:et:




import logging
import json
import requests
import time
import websocket

from platypush.config import Config
from platypush.message import Message

from .. import Backend

class PushbulletBackend(Backend):
    def __init__(self, token, device, **kwargs):
        super().__init__(**kwargs)

        self.token = token
        self.device_name = device
        self.pb_device_id = self.get_device_id()

        self._last_received_msg = {
            'request'  : { 'body': None, 'time': None },
            'response' : { 'body': None, 'time': None },
            'event'    : { 'body': None, 'time': None },
        }

    def _get_latest_push(self):
        t = int(time.time()) - 2
        try:
            response = requests.get(
                u'https://api.pushbullet.com/v2/pushes',
                headers = { 'Access-Token': self.token },
                params  = {
                    'modified_after': str(t),
                    'active' : 'true',
                    'limit'  : 1,
                }
            )

            response = response.json()
        except Exception as e:
            logging.exception(e)
            raise e

        if 'pushes' in response and response['pushes']:
            return response['pushes'][0]
        else:
            return {}

    def _should_skip_last_received_msg(self, msg):
        is_duplicate=False
        last_msg = self._last_received_msg[msg['type']]

        if last_msg:
            msg = Message.parse(msg)
            if str(msg) == str(last_msg['body']) \
                    and time.time() - last_msg['time'] <= 2:
                # Duplicate message sent on the Pushbullet socket within
                # two seconds, ignore it
                logging.debug('Ignoring duplicate message received on the socket')
                is_duplicate = True

        self._last_received_msg[msg['type']] = {
            'body': msg, 'time': time.time()
        }

        return is_duplicate

    def on_push(self):
        def _f(ws, data):
            try:
                try:
                    data = json.loads(data) if isinstance(data, str) else push
                except Exception as e:
                    logging.exception(e)
                    return

                if data['type'] == 'tickle' and data['subtype'] == 'push':
                    push = self._get_latest_push()
                elif data['type'] == 'push':
                    push = data['push']
                else: return  # Not a push notification

                if 'body' not in push: return
                logging.debug('Received push: {}'.format(push))

                body = push['body']
                try: body = json.loads(body)
                except ValueError as e: return  # Some other non-JSON push

                if not self._should_skip_last_received_msg(body):
                    self.on_message(body)
            except Exception as e:
                logging.exception(e)
                return

        return _f

    def on_error(self):
        def _f(ws, e):
            logging.exception(e)
            logging.info('Restarting PushBullet backend')
            ws.close()
            self._init_socket()

        return _f

    def _init_socket(self):
        self.ws = websocket.WebSocketApp(
            'wss://stream.pushbullet.com/websocket/' + self.token,
            on_message = self.on_push(),
            on_error = self.on_error())

    def get_device_id(self):
        response = requests.get(
            u'https://api.pushbullet.com/v2/devices',
            headers = { 'Access-Token': self.token },
        ).json()

        devices = [dev for dev in response['devices'] if 'nickname' in dev
                   and dev['nickname'] == self.device_name]

        if not devices:
            raise RuntimeError('No such Pushbullet device: {}'
                               .format(self.device_name))

        return devices[0]['iden']

    def send_message(self, msg):
        requests.post(
            u'https://api.pushbullet.com/v2/pushes',
            headers = { 'Access-Token': self.token },
            json = {
                'type': 'note',
                'device_iden': self.pb_device_id,
                'body': str(msg)
            }
        ).json()

    def on_stop(self):
        self.ws.close()

    def run(self):
        super().run()

        self._init_socket()
        logging.info('Initialized Pushbullet backend - device_id: {}'
                     .format(self.device_name))

        self.ws.run_forever()

# vim:sw=4:ts=4:et:




import json
import random
import re
import threading

from platypush.config import Config
from platypush.message import Message
from platypush.utils import get_event_class_by_type

class Event(Message):
    """ Event message class """

    def __init__(self, target=None, origin=None, id=None, **kwargs):
        """
        Params:
            target  -- Target node [String]
            origin  -- Origin node (default: current node) [String]
            id      -- Event ID (default: auto-generated)
            kwargs  -- Additional arguments for the event [kwDict]
        """

        self.id = id if id else self._generate_id()
        self.target = target if target else Config.get('device_id')
        self.origin = origin if origin else Config.get('device_id')
        self.type = '{}.{}'.format(self.__class__.__module__,
                                   self.__class__.__name__)
        self.args = kwargs

    @classmethod
    def build(cls, msg):
        """ Builds an event message from a JSON UTF-8 string/bytearray, a
        dictionary, or another Event """

        msg = super().parse(msg)
        event_type = msg['args'].pop('type')
        event_class = get_event_class_by_type(event_type)

        args = msg['args'] if 'args' in msg else {}
        args['id'] = msg['id'] if 'id' in msg else cls._generate_id()
        args['target'] = msg['target'] if 'target' in msg else Config.get('device_id')
        args['origin'] = msg['origin'] if 'origin' in msg else Config.get('device_id')
        return event_class(**args)


    @staticmethod
    def _generate_id():
        """ Generate a unique event ID """
        id = ''
        for i in range(0,16):
            id += '%.2x' % random.randint(0, 255)
        return id


    def matches_condition(self, condition):
        """
        If the event matches an event condition, it will return an EventMatchResult
        Params:
            -- condition -- The platypush.event.hook.EventCondition object
        """

        result = EventMatchResult(is_match=False, parsed_args=self.args)
        match_scores = []

        if not isinstance(self, condition.type): return result

        for (attr, value) in condition.args.items():
            if attr not in self.args:
                return result

            if isinstance(self.args[attr], str):
                arg_result = self._matches_argument(argname=attr, condition_value=value)

                if arg_result.is_match:
                    match_scores.append(arg_result.score)
                    for (parsed_arg, parsed_value) in arg_result.parsed_args.items():
                        result.parsed_args[parsed_arg] = parsed_value
                else:
                    return result
            elif self.args[attr] != value:
                # TODO proper support for list and dictionary matches
                return result

        result.is_match = True
        if match_scores:
            result.score = sum(match_scores) / float(len(match_scores))

        return result


    def _matches_argument(self, argname, condition_value):
        """
        Returns an EventMatchResult if the event argument [argname] matches
        [condition_value].

        - Example:
            - self.args = {
                'phrase': 'Hey dude turn on the living room lights'
            }

            - self._matches_argument(argname='phrase', condition_value='Turn on the $lights lights')
              will return EventMatchResult(is_match=True, parsed_args={ 'lights': 'living room' })

            - self._matches_argument(argname='phrase', condition_value='Turn off the $lights lights')
              will return EventMatchResult(is_match=False, parsed_args={})
        """

        result = EventMatchResult(is_match=False)
        event_tokens = re.split('\s+', self.args[argname].strip().lower())
        condition_tokens = re.split('\s+', condition_value.strip().lower())

        while event_tokens and condition_tokens:
            event_token = event_tokens[0]
            condition_token = condition_tokens[0]

            if event_token == condition_token:
                event_tokens.pop(0)
                condition_tokens.pop(0)
                result.score += 1
            elif re.search(condition_token, event_token):
                # The only supported regex-match as of now is the equivalent of
                # the maybe operator.
                # e.g. "turn on (the)? lights" would match both "turn on the lights"
                # and "turn on lights". In such a case, we just consume the
                # condition token and proceed forward. TODO add a more
                # sophisticated regex-match handling
                condition_tokens.pop(0)
            else:
                m = re.match('[^\\\]*\$([\w\d_-]+)', condition_token)
                if m:
                    argname = m.group(1)
                    if argname not in result.parsed_args:
                        result.parsed_args[argname] = event_token
                        result.score += 1
                    else:
                        result.parsed_args[argname] += ' ' + event_token


                    if len(condition_tokens) == 1 \
                            or (len(event_tokens) > 1 and len(condition_tokens) > 1 \
                            and event_tokens[1] == condition_tokens[1]):
                        # Stop appending tokens to this argument, as the next
                        # condition will be satisfied as well
                        condition_tokens.pop(0)

                    event_tokens.pop(0)
                else:
                    result.score -= 1
                    event_tokens.pop(0)

        # It's a match if all the tokens in the condition string have been satisfied
        result.is_match = len(condition_tokens) == 0
        return result


    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps({
            'type'     : 'event',
            'target'   : self.target,
            'origin'   : self.origin if hasattr(self, 'origin') else None,
            'id'       : self.id if hasattr(self, 'id') else None,
            'args'     : {
                'type' : self.type,
                **self.args,
            },
        })


class EventMatchResult(object):
    """ When comparing an event against an event condition, you want to
        return this object. It contains the match status (True or False),
        any parsed arguments, and a match_score that identifies how "strong"
        the match is - in case of multiple event matches, the ones with the
        highest score will win """

    def __init__(self, is_match, score=0, parsed_args=None):
        self.is_match = is_match
        self.score = score
        self.parsed_args = {} if not parsed_args else parsed_args


# XXX Should be a stop Request, not an Event
class StopEvent(Event):
    """ StopEvent message. When received on a Bus, it will terminate the
    listening thread having the specified ID. Useful to keep listeners in
    sync and make them quit when the application terminates """

    def __init__(self, target, origin, thread_id, id=None, **kwargs):
        """ Constructor.
        Params:
            target    -- Target node
            origin    -- Origin node
            thread_id -- thread_iden() to be terminated if listening on the bus
            id        -- Event ID (default: auto-generated)
            kwargs    -- Extra key-value arguments
        """

        super().__init__(target=target, origin=origin, id=id,
                         thread_id=thread_id, **kwargs)

    def targets_me(self):
        """ Returns true if the stop event is for the current thread """
        return self.args['thread_id'] == threading.get_ident()


# vim:sw=4:ts=4:et:




import logging

from ..config import Config
from ..message.request import Request

class Procedure(object):
    """ Procedure class. A procedure is a pre-configured list of requests """

    def __init__(self, name, requests, backend=None):
        """
        Params:
            name     -- Procedure name
            requests -- List of platylist.message.request.Request objects
        """

        self.name     = name
        self.requests = requests
        self.backend  = backend

        for req in requests:
            req.backend = self.backend

    @classmethod
    def build(cls, name, requests, backend=None, id=None, **kwargs):
        reqs = []
        for request_config in requests:
            request_config['origin'] = Config.get('device_id')
            request_config['id'] = id
            if 'target' not in request_config:
                request_config['target'] = request_config['origin']

            request = Request.build(request_config)
            reqs.append(request)

        return cls(name=name, requests=reqs, backend=backend, **kwargs)

    def execute(self, n_tries=1):
        """
        Execute the requests in the procedure
        Params:
            n_tries -- Number of tries in case of failure before raising a RuntimeError
        """

        logging.info('Executing request {}'.format(self.name))
        for request in self.requests:
            request.execute(n_tries)


# vim:sw=4:ts=4:et:




from platypush.context import get_backend
from platypush.message.response import Response

from platypush.plugins import Plugin

class AssistantGooglePlugin(Plugin):
    def start_conversation(self):
        assistant = get_backend('assistant.google')
        assistant.start_conversation()
        return Response(output='', errors=[])

    def stop_conversation(self):
        assistant = get_backend('assistant.google')
        assistant.stop_conversation()
        return Response(output='', errors=[])

# vim:sw=4:ts=4:et:




import re
import subprocess

from omxplayer import OMXPlayer

from platypush.context import get_plugin
from platypush.message.response import Response

from .. import Plugin

class MediaCtrlPlugin(Plugin):
    """
    Wrapper plugin to control audio and video media.
    Examples of supported URL types:
        - file:///media/movies/Movie.mp4 [requires omxplayer Python support]
        - youtube:video:poAk9XgK7Cs [requires omxplayer+youtube-dl]
        - magnet:?torrent_magnet [requires torrentcast]
        - spotify:track:track_id [leverages plugins.music.mpd]
    """

    def __init__(self, omxplayer_args=[], torrentcast_port=9090, *args, **kwargs):
        self.omxplayer_args = omxplayer_args
        self.torrentcast_port = torrentcast_port
        self.url = None
        self.plugin = None

    @classmethod
    def _get_type_and_resource_by_url(cls, url):
        # MPD/Mopidy media (TODO support more mopidy types)
        m = re.match('^https://open.spotify.com/([^/]+)/(.*)', url)
        if m: url = 'spotify:{}:{}'.format(m.group(1), m.group(2))
        if url.startswith('spotify:') \
                or url.startswith('tunein:') \
                or url.startswith('soundcloud:'):
            return ('mpd', url)

        # YouTube video
        m = re.match('youtube:video:(.*)', url)
        if m: url = 'https://www.youtube.com/watch?v={}'.format(m.group(1))
        if url.startswith('https://www.youtube.com/watch?v='):
            return ('youtube:video', url)

        # Local media
        if url.startswith('file://'):
            m = re.match('^file://(.*)', url)
            return ('file', m.group(1))

        # URL to a .torrent media or Magnet link
        if url.startswith('magnet:') or url.endswith('.torrent'):
            return ('torrent', url)

        raise RuntimeError('Unknown URL type: {}'.format(url))


    def play(self, url):
        (type, resource) = self._get_type_and_resource_by_url(url)
        response = Response(output='', errors = [])

        if type == 'mpd':
            self.plugin = get_plugin('music.mpd')
        elif type == 'youtube:video' or type == 'file':
            self.plugin = get_plugin('video.omxplayer')
        elif type == 'torrent':
            self.plugin = get_plugin('video.torrentcast')

        self.url = resource
        return self.plugin.play(resource)

    def pause(self):
        if self.plugin: return self.plugin.pause()


    def stop(self):
        if self.plugin: return self.plugin.stop()


    def voldown(self):
        if self.plugin: return self.plugin.voldown()


    def volup(self):
        if self.plugin: return self.plugin.volup()


    def back(self):
        if self.plugin: return self.plugin.back()


    def forward(self):
        if self.plugin: return self.plugin.forward()


# vim:sw=4:ts=4:et:




import logging
import re

from platypush.context import get_backend
from platypush.message.event import Event, EventMatchResult

class AssistantEvent(Event):
    """ Base class for assistant events """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self._assistant = get_backend('assistant.google')
        except KeyError as e:
            logging.warning('google.assistant backend not configured/initialized')
            self._assistant = None


class ConversationStartEvent(AssistantEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ConversationEndEvent(AssistantEvent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SpeechRecognizedEvent(AssistantEvent):
    def __init__(self, phrase, *args, **kwargs):
        super().__init__(phrase=phrase, *args, **kwargs)
        self.recognized_phrase = phrase.strip().lower()

    def matches_condition(self, condition):
        result = super().matches_condition(condition)
        if result.is_match and self._assistant:
            self._assistant.stop_conversation()

        return result


# vim:sw=4:ts=4:et:




from .. import Plugin

class MusicPlugin(Plugin):
    def play(self):
        raise NotImplementedError()

    def pause(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def next(self):
        raise NotImplementedError()

    def previous(self):
        raise NotImplementedError()

    def setvol(self, vol):
        raise NotImplementedError()

    def add(self, content):
        raise NotImplementedError()

    def playlistadd(self, playlist):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()

    def status(self):
        raise NotImplementedError()


# vim:sw=4:ts=4:et:




#!/usr/bin/env python

import errno
import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def pkg_files(dir):
    paths = []
    for (path, dirs, files) in os.walk(dir):
        for file in files:
            paths.append(os.path.join('..', path, file))
    return paths

def create_etc_dir():
    path = '/etc/platypush'
    try:
        os.makedirs(path)
    except OSError as e:
        if isinstance(e, PermissionError):
            print('WARNING: Could not create /etc/platypush')
        elif e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

plugins = pkg_files('platypush/plugins')
backend = pkg_files('platypush/backend')
create_etc_dir()

setup(
    name = "platypush",
    version = "0.6.4",
    author = "Fabio Manganiello",
    author_email = "info@fabiomanganiello.com",
    description = ("Platypush service"),
    license = "MIT",
    python_requires = '>= 3',
    keywords = "pushbullet notifications automation",
    url = "https://github.com/BlackLight/platypush",
    packages = find_packages(),
    entry_points = {
        'console_scripts': [
            'platypush=platypush.__main__',
            'pusher=platypush.pusher.__main__',
        ],
    },
    data_files = [
        ('/etc/platypush', ['platypush/config.example.yaml'])
    ],
    long_description = read('README.md'),
    classifiers = [
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
    ],
    install_requires = [
        'pyyaml',
        'requires',
    ],
    extras_require = {
        'Support for Apache Kafka backend': ['kafka-python'],
        'Support for Pushbullet backend': ['requests', 'websocket-client'],
        'Support for HTTP backend': ['flask'],
        'Support for Philips Hue plugin': ['phue'],
        'Support for MPD/Mopidy music server plugin': ['python-mpd2'],
        'Support for Belkin WeMo Switch plugin': ['ouimeaux'],
        'Support for text2speech plugin': ['mplayer'],
        'Support for OMXPlayer plugin': ['omxplayer'],
        'Support for YouTube in the OMXPlayer plugin': ['youtube-dl'],
        'Support for Google Assistant': ['google-assistant-sdk[samples]'],
        # 'Support for Flic buttons': ['-e git+https://github.com/50ButtonsEach/fliclib-linux-hci']
    },
)




from platypush.message.event import Event, EventMatchResult


class FlicButtonEvent(Event):
    """ Flic button event """

    def __init__(self, btn_addr, sequence, *args, **kwargs):
        super().__init__(btn_addr=btn_addr, sequence=sequence, *args, **kwargs)


    def matches_condition(self, condition):
        result = EventMatchResult(is_match=False)

        if not isinstance(self, condition.type) \
                or self.args['btn_addr'] != condition.args['btn_addr']:
            return result

        cond_sequence = list(condition.args['sequence'])
        event_sequence = list(self.args['sequence'])

        while cond_sequence and event_sequence:
            cond_press = cond_sequence[0]
            event_press = event_sequence[0]

            if cond_press == event_press:
                cond_sequence.pop(0)
                result.score += 1

            event_sequence.pop(0)

        result.is_match = len(cond_sequence) == 0
        return result


# vim:sw=4:ts=4:et:




import json
import logging
import re
import subprocess

from dbus.exceptions import DBusException
from omxplayer import OMXPlayer

from platypush.message.response import Response

from .. import Plugin

class VideoOmxplayerPlugin(Plugin):
    def __init__(self, args=[], *argv, **kwargs):
        self.args = args
        self.player = None

    def play(self, resource):
        if resource.startswith('youtube:') \
                or resource.startswith('https://www.youtube.com/watch?v='):
            resource = self._get_youtube_content(resource)

        try:
            self.player = OMXPlayer(resource, args=self.args)
        except DBusException as e:
            logging.warning('DBus connection failed: you will probably not ' +
                            'be able to control the media')
            logging.exception(e)

        return self.status()

    def pause(self):
        if self.player: self.player.play_pause()

    def stop(self):
        if self.player:
            self.player.stop()
            self.player.quit()
            self.player = None
        return self.status()

    def voldown(self):
        if self.player:
            self.player.set_volume(max(-6000, self.player.volume()-1000))
        return self.status()

    def volup(self):
        if self.player:
            self.player.set_volume(min(0, self.player.volume()+1000))
        return self.status()

    def back(self):
        if self.player:
            self.player.seek(-30)
        return self.status()

    def forward(self):
        if self.player:
            self.player.seek(+30)
        return self.status()

    def status(self):
        if self.player:
            return Response(output=json.dumps({
                'source': self.player.get_source(),
                'status': self.player.playback_status(),
                'volume': self.player.volume(),
                'elapsed': self.player.position(),
            }))
        else:
            return Response(output=json.dumps({
                'status': 'Not initialized'
            }))

    @classmethod
    def _get_youtube_content(cls, url):
        m = re.match('youtube:video:(.*)', url)
        if m: url = 'https://www.youtube.com/watch?v={}'.format(m.group(1))

        proc = subprocess.Popen(['youtube-dl','-f','best', '-g', url],
                                stdout=subprocess.PIPE)

        return proc.stdout.read().decode("utf-8", "strict")[:-1]


# vim:sw=4:ts=4:et:




import urllib3
import urllib.request

from platypush.message.response import Response

from .. import Plugin

class VideoTorrentcastPlugin(Plugin):
    def __init__(self, server='localhost', port=9090, *args, **kwargs):
        self.server = server
        self.port = port

    def play(self, url):
        request = urllib.request.urlopen(
            'http://{}:{}/play/'.format(self.server, self.port),
            data=urllib.parse.urlencode({
                'url': resource
            })
        )

        return Response(output=request.read())

    def pause(self):
        http = urllib3.PoolManager()
        request = http.request('POST',
            'http://{}:{}/pause/'.format(self.server, self.port))

        return Response(output=request.read())

    def stop(self):
        http = urllib3.PoolManager()
        request = http.request('POST',
            'http://{}:{}/stop/'.format(self.server, self.port))

        return Response(output=request.read())

    def voldown(self): return Response(output='Unsupported method')
    def volup(self): return Response(output='Unsupported method')
    def back(self): return Response(output='Unsupported method')
    def forward(self): return Response(output='Unsupported method')


# vim:sw=4:ts=4:et:




import logging
import os
import socket
import yaml

""" Config singleton instance """
_default_config_instance = None

class Config(object):
    """
    Configuration base class
    Usage:
        - Initialize config from one of the default paths:
            Config.init()
        - Initialize config from a custom path
            Config.init(config_file_path)
        - Set a value
            Config.set('foo', 'bar')
        - Get a value
            Config.get('foo')
    """

    """
    Default config file locations:
        - $HOME/.config/platypush/config.yaml
        - /etc/platypush/config.yaml
    """
    _cfgfile_locations = [
        os.path.join(os.environ['HOME'], '.config', 'platypush', 'config.yaml'),
        os.path.join(os.sep, 'etc', 'platypush', 'config.yaml'),
    ]

    def __init__(self, cfgfile=None):
        """
        Constructor. Always use the class as a singleton (i.e. through
        Config.init), you won't probably need to call the constructor directly
        Params:
            cfgfile -- Config file path (default: retrieve the first
                       available location in _cfgfile_locations)
        """

        if cfgfile is None:
            cfgfile = self._get_default_cfgfile()

        if cfgfile is None:
            raise RuntimeError('No config file specified and nothing found in {}'
                               .format(self._cfgfile_locations))

        with open(cfgfile, 'r') as fp:
            self._config = yaml.load(fp)

        for section in self._config:
            if 'disabled' in self._config[section] \
                    and self._config[section]['disabled']:
                del self._config[section]

        if 'logging' not in self._config:
            self._config['logging'] = logging.INFO
        else:
            self._config['logging'] = getattr(logging, self._config['logging'].upper())

        if 'device_id' not in self._config:
            self._config['device_id'] = socket.gethostname()

        self.backends = {}
        self.plugins = {}
        self.event_hooks = {}
        self.procedures = {}
        self._init_components()


    def _init_components(self):
        for key in self._config.keys():
            if key.startswith('backend.'):
                backend_name = '.'.join(key.split('.')[1:])
                self.backends[backend_name] = self._config[key]
            elif key.startswith('event.hook.'):
                hook_name = '.'.join(key.split('.')[2:])
                self.event_hooks[hook_name] = self._config[key]
            elif key.startswith('procedure.'):
                procedure_name = '.'.join(key.split('.')[1:])
                self.procedures[procedure_name] = self._config[key]
            else:
                self.plugins[key] = self._config[key]

    @staticmethod
    def get_backends():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.backends

    @staticmethod
    def get_plugins():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.plugins

    @staticmethod
    def get_event_hooks():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.event_hooks

    @staticmethod
    def get_procedures():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.procedures

    @staticmethod
    def get_default_pusher_backend():
        """
        Gets the default pusher backend from the config
        """
        backends = [k for k in Config.get_backends().keys()
                    if 'pusher' in Config.get_backends()[k]
                    and Config.get_backends()[k]['pusher'] is True]

        return backends[0] if backends else None


    @classmethod
    def _get_default_cfgfile(cls):
        for location in cls._cfgfile_locations:
            if os.path.isfile(location): return location

    @staticmethod
    def init(cfgfile=None):
        """
        Initializes the config object singleton
        Params:
            cfgfile -- path to the config file - default: _cfgfile_locations
        """
        global _default_config_instance
        _default_config_instance = Config(cfgfile)

    @staticmethod
    def get(key):
        """
        Gets a config value
        Params:
            key -- Config key to get
        """
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance._config[key]

    @staticmethod
    def set(key, value):
        """
        Sets a config value
        Params:
            key   -- Config key to set
            value -- Value for key
        """
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        _default_config_instance._config[key] = key

# vim:sw=4:ts=4:et:

